import io

import pytest

from grunnur import API, Context, Device, cuda_api_id, opencl_api_id
from grunnur.api import all_api_ids
from grunnur.adapter_base import APIID
from grunnur.virtual_alloc import TrivialManager, ZeroOffsetManager
from grunnur.testing import (
    MockPyCUDA,
    MockPyOpenCL,
    MockBackendFactory,
)

# Cannot just use the plugin directly since it is loaded before the coverage plugin,
# and all the function definitions in all `grunnur` modules get marked as not covered.
from pytest_grunnur import (
    get_devices,
    get_multi_device_sets,
)
from pytest_grunnur.plugin import (
    api,
    platform,
    device,
    context,
    some_device,
    some_context,
    multi_device_set,
    multi_device_context,
)

from pytest_grunnur.plugin import pytest_addoption as grunnur_pytest_addoption
from pytest_grunnur.plugin import pytest_generate_tests as grunnur_pytest_generate_tests
from pytest_grunnur.plugin import pytest_report_header as grunnur_pytest_report_header


@pytest.fixture
def mock_backend_factory(request, monkeypatch):
    yield MockBackendFactory(monkeypatch)


def pytest_generate_tests(metafunc):
    grunnur_pytest_generate_tests(metafunc)


def pytest_addoption(parser):
    grunnur_pytest_addoption(parser)


def pytest_report_header(config):
    grunnur_pytest_report_header(config)



import pytest

from grunnur import API, Platform, Device, Context


def run_tests(options=[]):
    pytest.main(["-v", "--no-cov", "-m", "plugin_inner_test"] + options)


@pytest.mark.plugin_inner_test
def test_api_fixture(api):
    assert isinstance(api, API)


def test_api(mock_backend_factory, capsys):
    backend_pycuda = mock_backend_factory.mock_pycuda()
    backend_pycuda.add_devices(["Device1"])
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Bar", ["Device2"])

    run_tests(["-k", "test_api_fixture"])
    captured = capsys.readouterr()
    assert "device(cuda,0,0): nVidia CUDA, Device1" in captured.out
    assert "device(opencl,0,0): Bar, Device2" in captured.out
    assert "::test_api_fixture[cuda]" in captured.out
    assert "::test_api_fixture[opencl]" in captured.out

    run_tests(["--api=cuda", "-k", "test_api_fixture"])
    captured = capsys.readouterr()
    assert "device(cuda,0,0): nVidia CUDA, Device1" in captured.out
    assert "device(opencl,0,0): Bar, Device2" not in captured.out
    assert "::test_api_fixture[cuda]" in captured.out
    assert "::test_api_fixture[opencl]" not in captured.out

    run_tests(["--api=opencl", "-k", "test_api_fixture"])
    captured = capsys.readouterr()
    assert "device(cuda,0,0): nVidia CUDA, Device1" not in captured.out
    assert "device(opencl,0,0): Bar, Device2" in captured.out
    assert "::test_api_fixture[cuda]" not in captured.out
    assert "::test_api_fixture[opencl]" in captured.out


def test_no_api(mock_backend_factory, capsys):
    run_tests(["-k", "test_api_fixture"])
    captured = capsys.readouterr()
    assert "::test_api_fixture[no_api]" in captured.out


@pytest.mark.plugin_inner_test
def test_platform_fixture(platform):
    assert isinstance(platform, Platform)


def test_platform(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1"])
    backend_pyopencl.add_platform_with_devices("Bar", ["Device2"])

    run_tests(["-k", "test_platform_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,1,0): Bar, Device2" in captured.out
    assert "::test_platform_fixture[opencl,0]" in captured.out
    assert "::test_platform_fixture[opencl,1]" in captured.out

    run_tests(["--platform-include-mask=Bar", "-k", "test_platform_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" not in captured.out
    assert "device(opencl,1,0): Bar, Device2" in captured.out
    assert "::test_platform_fixture[opencl,0]" not in captured.out
    assert "::test_platform_fixture[opencl,1]" in captured.out

    run_tests(["--platform-exclude-mask=Bar", "-k", "test_platform_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,1,0): Bar, Device2" not in captured.out
    assert "::test_platform_fixture[opencl,0]" in captured.out
    assert "::test_platform_fixture[opencl,1]" not in captured.out


def test_no_platform(mock_backend_factory, capsys):
    run_tests(["-k", "test_platform_fixture"])
    captured = capsys.readouterr()
    assert "::test_platform_fixture[no_platform]" in captured.out


@pytest.mark.plugin_inner_test
def test_device_fixture(device):
    assert isinstance(device, Device)


def test_device(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1", "Device2"])
    backend_pyopencl.add_platform_with_devices("Bar", ["Device2", "Device3"])

    run_tests(["-k", "test_device_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device2" in captured.out
    assert "device(opencl,1,0): Bar, Device2" in captured.out
    assert "device(opencl,1,1): Bar, Device3" in captured.out
    assert "::test_device_fixture[opencl,0,0]" in captured.out
    assert "::test_device_fixture[opencl,0,1]" in captured.out
    assert "::test_device_fixture[opencl,1,0]" in captured.out
    assert "::test_device_fixture[opencl,1,1]" in captured.out

    run_tests(["--device-include-mask=Device2", "-k", "test_device_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" not in captured.out
    assert "device(opencl,0,1): Foo, Device2" in captured.out
    assert "device(opencl,1,0): Bar, Device2" in captured.out
    assert "device(opencl,1,1): Bar, Device3" not in captured.out
    assert "::test_device_fixture[opencl,0,0]" not in captured.out
    assert "::test_device_fixture[opencl,0,1]" in captured.out
    assert "::test_device_fixture[opencl,1,0]" in captured.out
    assert "::test_device_fixture[opencl,1,1]" not in captured.out

    run_tests(["--device-exclude-mask=Device2", "-k", "test_device_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device2" not in captured.out
    assert "device(opencl,1,0): Bar, Device2" not in captured.out
    assert "device(opencl,1,1): Bar, Device3" in captured.out
    assert "::test_device_fixture[opencl,0,0]" in captured.out
    assert "::test_device_fixture[opencl,0,1]" not in captured.out
    assert "::test_device_fixture[opencl,1,0]" not in captured.out
    assert "::test_device_fixture[opencl,1,1]" in captured.out


@pytest.mark.plugin_inner_test
def test_some_device_fixture(some_device):
    assert isinstance(some_device, Device)


def test_some_device(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1", "Device2"])
    backend_pyopencl.add_platform_with_devices("Bar", ["Device2", "Device3"])

    run_tests(["-k", "test_some_device_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device2" in captured.out
    assert "device(opencl,1,0): Bar, Device2" in captured.out
    assert "device(opencl,1,1): Bar, Device3" in captured.out
    assert "::test_some_device_fixture[opencl,0,0]" in captured.out
    assert "::test_some_device_fixture[opencl,0,1]" not in captured.out
    assert "::test_some_device_fixture[opencl,1,0]" not in captured.out
    assert "::test_some_device_fixture[opencl,1,1]" not in captured.out

    run_tests(["--device-include-mask=Device2", "-k", "test_some_device_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" not in captured.out
    assert "device(opencl,0,1): Foo, Device2" in captured.out
    assert "device(opencl,1,0): Bar, Device2" in captured.out
    assert "device(opencl,1,1): Bar, Device3" not in captured.out
    assert "::test_some_device_fixture[opencl,0,0]" not in captured.out
    assert "::test_some_device_fixture[opencl,0,1]" in captured.out
    assert "::test_some_device_fixture[opencl,1,0]" not in captured.out
    assert "::test_some_device_fixture[opencl,1,1]" not in captured.out


def test_duplicate_devices(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1", "Device1", "Device2"])

    run_tests(["-k", "test_device_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device1" not in captured.out
    assert "device(opencl,0,2): Foo, Device2" in captured.out
    assert "::test_device_fixture[opencl,0,0]" in captured.out
    assert "::test_device_fixture[opencl,0,1]" not in captured.out
    assert "::test_device_fixture[opencl,0,2]" in captured.out

    run_tests(["--include-duplicate-devices", "-k", "test_device_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device1" in captured.out
    assert "device(opencl,0,2): Foo, Device2" in captured.out
    assert "::test_device_fixture[opencl,0,0]" in captured.out
    assert "::test_device_fixture[opencl,0,1]" in captured.out
    assert "::test_device_fixture[opencl,0,2]" in captured.out


def test_no_device(mock_backend_factory, capsys):
    run_tests(["-k", "test_device_fixture"])
    captured = capsys.readouterr()
    assert "No GPGPU devices available" in captured.out
    assert "::test_device_fixture[no_device]" in captured.out


@pytest.mark.plugin_inner_test
def test_context_fixture(context):
    assert isinstance(context, Context)
    assert len(context.devices) == 1


def test_context(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1", "Device2"])

    run_tests(["-k", "test_context_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device2" in captured.out
    assert "::test_context_fixture[opencl,0,0]" in captured.out
    assert "::test_context_fixture[opencl,0,1]" in captured.out

    run_tests(["--device-include-mask=Device1", "-k", "test_context_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device2" not in captured.out
    assert "::test_context_fixture[opencl,0,0]" in captured.out
    assert "::test_context_fixture[opencl,0,1]" not in captured.out


@pytest.mark.plugin_inner_test
def test_some_context_fixture(some_context):
    assert isinstance(some_context, Context)
    assert len(some_context.devices) == 1


def test_some_context(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1", "Device2"])

    run_tests(["-k", "test_some_context_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    assert "device(opencl,0,1): Foo, Device2" in captured.out
    assert "::test_some_context_fixture[opencl,0,0]" in captured.out
    assert "::test_some_context_fixture[opencl,0,1]" not in captured.out

    run_tests(["--device-include-mask=Device2", "-k", "test_some_context_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" not in captured.out
    assert "device(opencl,0,1): Foo, Device2" in captured.out
    assert "::test_some_context_fixture[opencl,0,0]" not in captured.out
    assert "::test_some_context_fixture[opencl,0,1]" in captured.out


def test_no_context(mock_backend_factory, capsys):
    run_tests(["-k", "test_context_fixture"])
    captured = capsys.readouterr()
    assert "::test_context_fixture[no_device]" in captured.out


@pytest.mark.plugin_inner_test
def test_multi_device_context_fixture(multi_device_context):
    assert isinstance(multi_device_context, Context)
    assert len(multi_device_context.devices) > 1


# FIXME: decide on the exact logic in this case.
def test_multi_device_context(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    # Two of the devices have the same names to check that they will be picked up
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1", "Device1", "Device3"])

    run_tests(["-k", "test_multi_device_context_fixture"])
    captured = capsys.readouterr()
    assert "device(opencl,0,0): Foo, Device1" in captured.out
    # Multi-device context does not currently include all the devices used to the list
    assert "device(opencl,0,1): Foo, Device1" not in captured.out
    assert "device(opencl,0,2): Foo, Device3" in captured.out
    assert "::test_multi_device_context_fixture[opencl,0,0+opencl,0,1+opencl,0,2]" in captured.out


def test_no_multi_device_context(mock_backend_factory, capsys):
    backend_pyopencl = mock_backend_factory.mock_pyopencl()
    backend_pyopencl.add_platform_with_devices("Foo", ["Device1"])

    run_tests(["-k", "test_multi_device_context_fixture"])
    captured = capsys.readouterr()
    assert "::test_multi_device_context_fixture[no_multi_device]" in captured.out



from functools import lru_cache
from typing import Iterable, TypeVar, Optional, List, Iterator, Any, Tuple, cast

import pytest

from grunnur import API, all_api_ids, Platform, PlatformFilter, Device, DeviceFilter, Context


def pytest_addoption(parser: pytest.Parser) -> None:
    """
    Adds the following command-line options:

    * ``--api``: select a specific API to test (out of returned by :py:func:`grunnur.all_api_ids`).
    * ``--platform-include-mask``: run tests only on platforms whose names matches the mask.
    * ``--platform-exclude-mask``: exclude platforms whose names matches the mask from the tests.
    * ``--device-include-mask``: run tests only on devices whose names matches the mask.
    * ``--device-exclude-mask``: exclude devices whose names matches the mask from the tests.
    * ``--include-duplicate-devices``: if there are devices with the same name in the platform,
      run tests on all of them.
    * ``--include-pure-parallel-devices``: include pure parallel devices in the tests
      (that is, those not supporting synchronization within a block/work group).
    """

    api_shortcuts = [api_id.shortcut for api_id in all_api_ids()]
    parser.addoption(
        "--api",
        action="store",
        help="GPGPU API: " + "/".join(api_shortcuts) + " (or all available if not given)",
        default=None,
        choices=api_shortcuts,
    )

    parser.addoption(
        "--platform-include-mask",
        action="append",
        help="Run tests on matching platforms only",
        default=[],
    )
    parser.addoption(
        "--platform-exclude-mask",
        action="append",
        help="Run tests on matching platforms only",
        default=[],
    )

    parser.addoption(
        "--device-include-mask",
        action="append",
        help="Run tests on matching devices only",
        default=[],
    )
    parser.addoption(
        "--device-exclude-mask",
        action="append",
        help="Run tests on matching devices only",
        default=[],
    )

    parser.addoption(
        "--include-duplicate-devices",
        action="store_true",
        help="Run tests on all available devices and not only on uniquely named ones",
        default=False,
    )

    parser.addoption(
        "--include-pure-parallel-devices",
        action="store_true",
        help="Include pure parallel devices (not supporting synchronization within a work group)",
        default=False,
    )


@lru_cache()
def get_apis(config: pytest.Config) -> List[API]:
    """
    Returns the list of APIs filtered by the test configuration.
    """
    return API.all_by_shortcut(config.option.api)


_T = TypeVar("_T")


def concatenate(lists: Iterable[List[_T]]) -> List[_T]:
    return sum(lists, [])


@lru_cache()
def get_platforms(config: pytest.Config) -> List[Platform]:
    """
    Returns the list of platforms filtered by the test configuration
    (concatenated for all filtered APIs).
    """
    apis = get_apis(config)
    return concatenate(
        Platform.all_filtered(
            api,
            PlatformFilter(
                include_masks=config.option.platform_include_mask,
                exclude_masks=config.option.platform_exclude_mask,
            ),
        )
        for api in apis
    )


@lru_cache()
def get_device_sets(
    config: pytest.Config, unique_devices_only_override: Optional[bool] = None
) -> List[List[Device]]:
    if unique_devices_only_override is not None:
        unique_devices_only = unique_devices_only_override
    else:
        unique_devices_only = not config.option.include_duplicate_devices

    platforms = get_platforms(config)
    return [
        Device.all_filtered(
            platform,
            DeviceFilter(
                include_masks=config.option.device_include_mask,
                exclude_masks=config.option.device_exclude_mask,
                unique_only=unique_devices_only,
                exclude_pure_parallel=not config.option.include_pure_parallel_devices,
            ),
        )
        for platform in platforms
    ]


@lru_cache()
def get_devices(config: pytest.Config) -> List[Device]:
    """
    Returns the list of devices filtered by the test configuration
    (concatenated for all filtered platforms and APIs).
    """
    return concatenate(get_device_sets(config))


@lru_cache()
def get_multi_device_sets(config: pytest.Config) -> List[List[Device]]:
    """
    Returns a list where each element is a list with two or more devices
    belonging to the same API and platform, where APIs, platforms, and devices
    are filtered by the test configuration.
    """
    device_sets = get_device_sets(config, unique_devices_only_override=False)
    return [device_set for device_set in device_sets if len(device_set) > 1]


@pytest.fixture
def api(request: pytest.FixtureRequest) -> Iterator[API]:
    """
    Yields the elements of the return value of :py:func:`~pytest_grunnur.get_apis`.
    """
    yield request.param


@pytest.fixture
def platform(request: pytest.FixtureRequest) -> Iterator[Platform]:
    """
    Yields the elements of the return value of :py:func:`~pytest_grunnur.get_platforms`.
    """
    yield request.param


@pytest.fixture
def device(request: pytest.FixtureRequest) -> Iterator[Device]:
    """
    Yields the elements of the return value of :py:func:`~pytest_grunnur.get_devices`.
    """
    yield request.param


@pytest.fixture
def context(device: Device) -> Iterator[Context]:
    """
    A single-device context for each device yielded by :py:func:`~pytest_grunnur.plugin.device`.
    """
    yield Context.from_devices([device])


@pytest.fixture
def some_device(request: pytest.FixtureRequest) -> Iterator[Device]:
    """
    Yields one element of the return value of :py:func:`~pytest_grunnur.get_devices`.
    """
    yield request.param


@pytest.fixture
def some_context(some_device: Device) -> Iterator[Context]:
    """
    A single-device context initialized with the return value of
    :py:func:`~pytest_grunnur.plugin.some_device`.
    """
    yield Context.from_devices([some_device])


@pytest.fixture
def multi_device_set(request: pytest.FixtureRequest) -> Iterator[List[Device]]:
    """
    Yields the elements of the return value of :py:func:`~pytest_grunnur.get_multi_device_sets`.
    """
    yield request.param


@pytest.fixture
def multi_device_context(multi_device_set: List[Device]) -> Iterator[Context]:
    """
    A multi-device context for each device set yielded by
    :py:func:`~pytest_grunnur.plugin.multi_device_set`.
    """
    yield Context.from_devices(multi_device_set)


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """
    Seeds the parameters for the fixtures provided by this plugin
    (see the fixture list for details).
    """

    apis = get_apis(metafunc.config)
    platforms = get_platforms(metafunc.config)
    devices = get_devices(metafunc.config)

    fixtures: List[Tuple[str, List[Any]]] = [
        ("api", apis),
        ("platform", platforms),
        ("device", devices),
    ]

    for name, vals in fixtures:
        if name in metafunc.fixturenames:
            metafunc.parametrize(
                name,
                vals,
                ids=["no_" + name] if len(vals) == 0 else lambda obj: cast(str, obj.shortcut),
                indirect=True,
            )

    if "some_device" in metafunc.fixturenames:
        metafunc.parametrize(
            "some_device",
            devices if len(devices) == 0 else [devices[0]],
            ids=["no_device"] if len(devices) == 0 else lambda device: cast(str, device.shortcut),
            indirect=True,
        )

    if "multi_device_set" in metafunc.fixturenames:
        device_sets = get_multi_device_sets(metafunc.config)
        ids = ["+".join(device.shortcut for device in device_set) for device_set in device_sets]
        metafunc.parametrize(
            "multi_device_set",
            device_sets,
            ids=["no_multi_device"] if len(device_sets) == 0 else ids,
            indirect=True,
        )


def pytest_report_header(config: pytest.Config) -> None:
    """
    Adds a header to the test report, listing all the GPGPU devices the tests are run on,
    including their short numerical IDs (appearing in the test parameters).
    """
    devices = get_devices(config)

    if len(devices) == 0:
        print("No GPGPU devices available")
    else:
        print("Running tests on:")
        for device in sorted(devices, key=lambda device: str(device)):
            platform = device.platform
            print(f"  {device}: {platform.name}, {device.name}")
